// Copyright The OpenTelemetry Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package hostmetricsreceiver

import (
	"context"
	"errors"
	"runtime"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"go.uber.org/zap"

	"go.opentelemetry.io/collector/component/componenttest"
	"go.opentelemetry.io/collector/consumer/pdata"
	"go.opentelemetry.io/collector/consumer/pdatautil"
	"go.opentelemetry.io/collector/exporter/exportertest"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/cpuscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/diskscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/filesystemscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/loadscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/memoryscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/networkscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/processscraper"
	"go.opentelemetry.io/collector/receiver/hostmetricsreceiver/internal/scraper/virtualmemoryscraper"
)

var standardMetrics = []string{
	"host/cpu/usage",
	"host/memory/used",
	"host/disk/bytes",
	"host/disk/ops",
	"host/disk/time",
	"host/filesystem/used",
	"host/load/1m",
	"host/load/5m",
	"host/load/15m",
	"host/network/packets",
	"host/network/dropped_packets",
	"host/network/errors",
	"host/network/bytes",
	"host/network/tcp_connections",
	"host/swap/paging",
	"host/swap/usage",
}

var resourceMetrics = []string{
	"process/cpu/usage",
	"process/memory/usage",
	"process/disk/bytes",
}

var systemSpecificMetrics = map[string][]string{
	"linux":   {"host/filesystem/inodes/used", "host/swap/page_faults"},
	"darwin":  {"host/filesystem/inodes/used", "host/swap/page_faults"},
	"freebsd": {"host/filesystem/inodes/used", "host/swap/page_faults"},
	"openbsd": {"host/filesystem/inodes/used", "host/swap/page_faults"},
	"solaris": {"host/filesystem/inodes/used", "host/swap/page_faults"},
}

var factories = map[string]internal.ScraperFactory{
	cpuscraper.TypeStr:           &cpuscraper.Factory{},
	diskscraper.TypeStr:          &diskscraper.Factory{},
	filesystemscraper.TypeStr:    &filesystemscraper.Factory{},
	loadscraper.TypeStr:          &loadscraper.Factory{},
	memoryscraper.TypeStr:        &memoryscraper.Factory{},
	networkscraper.TypeStr:       &networkscraper.Factory{},
	virtualmemoryscraper.TypeStr: &virtualmemoryscraper.Factory{},
}

var resourceFactories = map[string]internal.ResourceScraperFactory{
	processscraper.TypeStr: &processscraper.Factory{},
}

func TestGatherMetrics_EndToEnd(t *testing.T) {
	sink := &exportertest.SinkMetricsExporter{}

	config := &Config{
		CollectionInterval: 100 * time.Millisecond,
		Scrapers: map[string]internal.Config{
			cpuscraper.TypeStr:           &cpuscraper.Config{ReportPerCPU: true},
			diskscraper.TypeStr:          &diskscraper.Config{},
			filesystemscraper.TypeStr:    &filesystemscraper.Config{},
			loadscraper.TypeStr:          &loadscraper.Config{},
			memoryscraper.TypeStr:        &memoryscraper.Config{},
			networkscraper.TypeStr:       &networkscraper.Config{},
			processscraper.TypeStr:       &processscraper.Config{},
			virtualmemoryscraper.TypeStr: &virtualmemoryscraper.Config{},
		},
	}

	receiver, err := newHostMetricsReceiver(context.Background(), zap.NewNop(), config, factories, resourceFactories, sink)

	require.NoError(t, err, "Failed to create metrics receiver: %v", err)

	ctx, cancelFn := context.WithCancel(context.Background())
	err = receiver.Start(ctx, componenttest.NewNopHost())
	require.NoError(t, err, "Failed to start metrics receiver: %v", err)
	defer func() { assert.NoError(t, receiver.Shutdown(context.Background())) }()

	// canceling the context provided to Start should not cancel any async processes initiated by the receiver
	cancelFn()

	require.Eventually(t, func() bool {
		got := sink.AllMetrics()
		if len(got) == 0 {
			return false
		}

		assertIncludesStandardMetrics(t, got[0])
		assertIncludesResourceMetrics(t, got[0])
		return true
	}, time.Second, 10*time.Millisecond, "No metrics were collected after 1s")
}

func assertIncludesStandardMetrics(t *testing.T, got pdata.Metrics) {
	md := pdatautil.MetricsToInternalMetrics(got)

	// get the first ResourceMetrics object
	rms := md.ResourceMetrics()
	require.GreaterOrEqual(t, rms.Len(), 1)
	rm := rms.At(0)
	assert.True(t, rm.Resource().IsNil() || rm.Resource().Attributes().Len() == 0)

	metrics := getMetricSlice(t, rm)
	returnedMetrics := getReturnedMetricNames(metrics)

	// the expected list of metrics returned is os dependent
	expectedMetrics := append(standardMetrics, systemSpecificMetrics[runtime.GOOS]...)
	assert.Equal(t, len(expectedMetrics), len(returnedMetrics))
	for _, expected := range expectedMetrics {
		assert.Contains(t, returnedMetrics, expected)
	}
}

func assertIncludesResourceMetrics(t *testing.T, got pdata.Metrics) {
	md := pdatautil.MetricsToInternalMetrics(got)

	// get the superset of metrics returned by all resource metrics (excluding the first)
	returnedMetrics := make(map[string]struct{})
	rms := md.ResourceMetrics()
	for i := 1; i < rms.Len(); i++ {
		rm := rms.At(i)
		assert.Greater(t, rm.Resource().Attributes().Len(), 0)
		metrics := getMetricSlice(t, rm)
		appendMapInto(returnedMetrics, getReturnedMetricNames(metrics))
	}

	assert.Equal(t, len(resourceMetrics), len(returnedMetrics))
	for _, expected := range resourceMetrics {
		assert.Contains(t, returnedMetrics, expected)
	}
}

func getMetricSlice(t *testing.T, rm pdata.ResourceMetrics) pdata.MetricSlice {
	ilms := rm.InstrumentationLibraryMetrics()
	require.Equal(t, 1, ilms.Len())
	return ilms.At(0).Metrics()
}

func getReturnedMetricNames(metrics pdata.MetricSlice) map[string]struct{} {
	metricNames := make(map[string]struct{})
	for i := 0; i < metrics.Len(); i++ {
		metricNames[metrics.At(i).MetricDescriptor().Name()] = struct{}{}
	}
	return metricNames
}

func appendMapInto(m1 map[string]struct{}, m2 map[string]struct{}) {
	for k, v := range m2 {
		m1[k] = v
	}
}

const mockTypeStr = "mock"
const mockResourceTypeStr = "mockresource"

type mockConfig struct{}

type mockFactory struct{}
type mockScraper struct{}

func (*mockFactory) CreateDefaultConfig() internal.Config { return &mockConfig{} }
func (*mockFactory) CreateMetricsScraper(ctx context.Context, logger *zap.Logger, cfg internal.Config) (internal.Scraper, error) {
	return &mockScraper{}, nil
}

func (*mockScraper) Initialize(ctx context.Context) error { return nil }
func (*mockScraper) Close(ctx context.Context) error      { return nil }
func (*mockScraper) ScrapeMetrics(ctx context.Context) (pdata.MetricSlice, error) {
	return pdata.NewMetricSlice(), errors.New("err1")
}

type mockResourceFactory struct{}
type mockResourceScraper struct{}

func (*mockResourceFactory) CreateDefaultConfig() internal.Config { return &mockConfig{} }
func (*mockResourceFactory) CreateMetricsScraper(ctx context.Context, logger *zap.Logger, cfg internal.Config) (internal.ResourceScraper, error) {
	return &mockResourceScraper{}, nil
}

func (*mockResourceScraper) Initialize(ctx context.Context) error { return nil }
func (*mockResourceScraper) Close(ctx context.Context) error      { return nil }
func (*mockResourceScraper) ScrapeMetrics(ctx context.Context) (pdata.ResourceMetricsSlice, error) {
	return pdata.NewResourceMetricsSlice(), errors.New("err2")
}

func TestGatherMetrics_ScraperKeyConfigError(t *testing.T) {
	var mockFactories = map[string]internal.ScraperFactory{}
	var mockResourceFactories = map[string]internal.ResourceScraperFactory{}

	sink := &exportertest.SinkMetricsExporter{}
	config := &Config{Scrapers: map[string]internal.Config{"error": &mockConfig{}}}

	_, err := newHostMetricsReceiver(context.Background(), zap.NewNop(), config, mockFactories, mockResourceFactories, sink)
	require.Error(t, err)
}

func TestGatherMetrics_Error(t *testing.T) {
	var mockFactories = map[string]internal.ScraperFactory{mockTypeStr: &mockFactory{}}
	var mockResourceFactories = map[string]internal.ResourceScraperFactory{mockResourceTypeStr: &mockResourceFactory{}}

	sink := &exportertest.SinkMetricsExporter{}

	config := &Config{
		Scrapers: map[string]internal.Config{
			mockTypeStr:         &mockConfig{},
			mockResourceTypeStr: &mockConfig{},
		},
	}

	receiver, err := newHostMetricsReceiver(context.Background(), zap.NewNop(), config, mockFactories, mockResourceFactories, sink)
	require.NoError(t, err)

	receiver.initializeScrapers(context.Background(), componenttest.NewNopHost())
	receiver.scrapeMetrics(context.Background())

	got := sink.AllMetrics()

	// expect to get one empty resource metrics entry
	require.Equal(t, 1, len(got))
	rm := pdatautil.MetricsToInternalMetrics(got[0]).ResourceMetrics()
	require.Equal(t, 1, rm.Len())
	ilm := rm.At(0).InstrumentationLibraryMetrics()
	require.Equal(t, 1, ilm.Len())
	metrics := ilm.At(0).Metrics()
	require.Equal(t, 0, metrics.Len())
}

func benchmarkScrapeMetrics(b *testing.B, cfg *Config) {
	sink := &exportertest.SinkMetricsExporter{}

	receiver, _ := newHostMetricsReceiver(context.Background(), zap.NewNop(), cfg, factories, resourceFactories, sink)
	receiver.initializeScrapers(context.Background(), componenttest.NewNopHost())

	b.ResetTimer()
	for n := 0; n < b.N; n++ {
		receiver.scrapeMetrics(context.Background())
	}

	if len(sink.AllMetrics()) == 0 {
		b.Fail()
	}
}

func Benchmark_ScrapeCpuMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{cpuscraper.TypeStr: (&cpuscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeDiskMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{diskscraper.TypeStr: (&diskscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeFileSystemMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{filesystemscraper.TypeStr: (&filesystemscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeLoadMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{loadscraper.TypeStr: (&loadscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeMemoryMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{memoryscraper.TypeStr: (&memoryscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeNetworkMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{networkscraper.TypeStr: (&networkscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeProcessMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{processscraper.TypeStr: (&processscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeVirtualMemoryMetrics(b *testing.B) {
	cfg := &Config{Scrapers: map[string]internal.Config{virtualmemoryscraper.TypeStr: (&virtualmemoryscraper.Factory{}).CreateDefaultConfig()}}
	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeDefaultMetrics(b *testing.B) {
	cfg := &Config{
		Scrapers: map[string]internal.Config{
			cpuscraper.TypeStr:           (&cpuscraper.Factory{}).CreateDefaultConfig(),
			diskscraper.TypeStr:          (&diskscraper.Factory{}).CreateDefaultConfig(),
			filesystemscraper.TypeStr:    (&filesystemscraper.Factory{}).CreateDefaultConfig(),
			loadscraper.TypeStr:          (&loadscraper.Factory{}).CreateDefaultConfig(),
			memoryscraper.TypeStr:        (&memoryscraper.Factory{}).CreateDefaultConfig(),
			networkscraper.TypeStr:       (&networkscraper.Factory{}).CreateDefaultConfig(),
			virtualmemoryscraper.TypeStr: (&virtualmemoryscraper.Factory{}).CreateDefaultConfig(),
		},
	}

	benchmarkScrapeMetrics(b, cfg)
}

func Benchmark_ScrapeAllMetrics(b *testing.B) {
	cfg := &Config{
		Scrapers: map[string]internal.Config{
			cpuscraper.TypeStr:           &cpuscraper.Config{ReportPerCPU: true},
			diskscraper.TypeStr:          &diskscraper.Config{},
			filesystemscraper.TypeStr:    &filesystemscraper.Config{},
			loadscraper.TypeStr:          &loadscraper.Config{},
			memoryscraper.TypeStr:        &memoryscraper.Config{},
			networkscraper.TypeStr:       &networkscraper.Config{},
			processscraper.TypeStr:       &processscraper.Config{},
			virtualmemoryscraper.TypeStr: &virtualmemoryscraper.Config{},
		},
	}

	benchmarkScrapeMetrics(b, cfg)
}
